"""
The `hydra-train` creates an AllenNLP config by composing a set of `yaml` files
with Hydra's [`Compose API`](https://hydra.cc/docs/advanced/compose_api) then
passing it to AllenNLP's `train` method.

Overriding using Hydra's [`Override Grammar`](https://hydra.cc/docs/advanced/compose_api)
is also supported.

The `hydra-train` command uses the underlying functionality of the
[`compose`](/allennlp-hydra/site/hydra/commands/compose_config) command. So the
use that documentation for understanding the composition of `.yaml` files.

# Parameters

config_path: `Union[str, PathLike]`
    Path to the root config directory.

config_name: `str`
    The name of the root config file. Do NOT include the `.yaml`.

job_name: `str`
    The job name. This is passed to Hydra and is not used here.

-s/--serialization-dir: `Union[str, PathLike]`
    The directory where everything is saved.

-r/--recover: `bool`, optional (default=`False`)
    Flag. Recover training from the state in `serialization_dir`

-f/--force: `bool`, optional (default=`False`)
    Flag. Overwrite the output directory if it exists.

--node-rank: `int`, optional (default=`0`)
    The rank of this node in the distributed setup

--dry-run: `bool`, optional (default=`False`)
    Flag. Do not train the model, but create a vocabulary, show dataset statistics
    and other training information

--file-friendly-logging: `bool`, optional (default=`False`)
    Flag. Outputs tqdm status on separate lines and slows tqdm refresh rate

-o/--overrides: `List[str]`, optional (default=`[]`)
    Keyword arguments passed will be used as a list of overrides using Hydra's
    override grammar for the config.

    Example usage:

    ```zsh
    --overrides A=B C="D"
    ```
    Will be interpreted as overrides `['A=B', 'C="D"']`
"""

import argparse
import logging

from overrides import overrides

from allennlp.commands.subcommand import Subcommand
from allennlp.commands.train import train_model
from allennlp.common import Params

logger = logging.getLogger(__name__)


@Subcommand.register("hydra-train")
class HydraTrain(Subcommand):
    @overrides
    def add_subparser(
        self, parser: argparse._SubParsersAction
    ) -> argparse.ArgumentParser:
        description = """
        Train the specified model on the specified dataset with a hydra config.
        """
        subparser = parser.add_parser(
            self.name, description=description, help="Train a model."
        )

        subparser.add_argument(
            "config_path", type=str, help="Path to the config directory."
        )

        subparser.add_argument(
            "config_name", type=str, help="Name of the config file to use."
        )
        subparser.add_argument("job_name", type=str, help="Name of the job.")

        subparser.add_argument(
            "-s",
            "--serialization-dir",
            required=True,
            type=str,
            help="directory in which to save the model and its logs",
        )

        subparser.add_argument(
            "-r",
            "--recover",
            action="store_true",
            default=False,
            help="recover training from the state in serialization_dir",
        )

        subparser.add_argument(
            "-f",
            "--force",
            action="store_true",
            required=False,
            help="overwrite the output directory if it exists",
        )

        subparser.add_argument(
            "--node-rank",
            type=int,
            default=0,
            help="rank of this node in the distributed setup",
        )

        subparser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help=(
                "do not train a model, but create a vocabulary, show dataset statistics and "
                "other training information"
            ),
        )
        subparser.add_argument(
            "--file-friendly-logging",
            action="store_true",
            default=False,
            help="outputs tqdm status on separate lines and slows tqdm refresh rate",
        )
        subparser.add_argument(
            "-o",
            "--overrides",
            nargs="*",
            help="Any key=value arguments to override config values "
            "(use dots for.nested=overrides)",
        )

        subparser.add_argument(
            "--fill-defaults",
            action="store_true",
            default=False,
            help="Add default arguments from each loaded class to the config.",
        )

        subparser.set_defaults(func=hydra_train_model_from_args)

        return subparser


def hydra_train_model_from_args(args: argparse.Namespace):
    """
    Just converts from an `argparse.Namespace` object to string paths.
    """

    # Load the hydra config, overrides will be used here.
    from allennlp_hydra.commands import compose_config

    # We do NOT pass a serialization dir to the compose because we do not want
    # to save the config here. `train_model` handles that for us.
    config = compose_config.compose_config(
        config_path=args.config_path,
        config_name=args.config_name,
        job_name=args.job_name,
        serialization_dir=None,
        config_overrides=args.overrides,
        fill_defaults=args.fill_defaults,
    )

    return train_model(
        params=Params(config),
        serialization_dir=args.serialization_dir,
        recover=args.recover,
        force=args.force,
        node_rank=args.node_rank,
        include_package=args.include_package,
        dry_run=args.dry_run,
        file_friendly_logging=args.file_friendly_logging,
    )



import pathlib
import pytest
import logging

__all__ = ["PROJECT_ROOT", "TEST_ROOT", "FIXTURES_ROOT", "BaseTestCase"]

PROJECT_ROOT = pathlib.Path(__file__).parents[3].resolve()
TEST_ROOT = PROJECT_ROOT.joinpath("tests")
FIXTURES_ROOT = PROJECT_ROOT.joinpath("test_fixtures")


class BaseTestCase:
    """
    A custom testing class that disables some of the more verbose AllenNLP
    logging and that creates and destroys a temp directory as a test fixture.
    """

    PROJECT_ROOT = PROJECT_ROOT
    MODULE_ROOT = PROJECT_ROOT / "src"
    TESTS_ROOT = TEST_ROOT
    FIXTURES_ROOT = FIXTURES_ROOT
    FIXTURES_DATA_PATH = FIXTURES_ROOT / "data"

    # This is crucial for making the test cases not have permission denied
    # issues later down the line.
    @pytest.fixture(autouse=True)
    def test_dir(self, tmpdir):
        self.TEST_DIR = pathlib.Path(tmpdir.mkdir("allennlp_hydra"))

    def setup_method(self):
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
            level=logging.DEBUG,
        )
        # Disabling some of the more verbose logging statements that typically aren't very helpful
        # in tests.
        logging.getLogger("allennlp.common.params").disabled = True
        logging.getLogger("allennlp.nn.initializers").disabled = True
        logging.getLogger("allennlp.modules.token_embedders.embedding").setLevel(
            logging.INFO
        )
        logging.getLogger("urllib3.connectionpool").disabled = True




from allennlp_hydra.utils.testing.test_case import (
    PROJECT_ROOT,
    TEST_ROOT,
    FIXTURES_ROOT,
    BaseTestCase,
)
from allennlp_hydra.utils.testing.utils import assert_models_weights_equal



from numpy.testing import assert_allclose


def assert_models_weights_equal(model_1, model_2):
    model_1_state_keys = model_1.state_dict().keys()
    model_2_state_keys = model_2.state_dict().keys()
    assert model_1_state_keys == model_2_state_keys
    for key in model_1_state_keys:
        assert_allclose(
            model_1.state_dict()[key].cpu().numpy(),
            model_2.state_dict()[key].cpu().numpy(),
            err_msg=key,
        )



"""
Command to convert a class to a yaml. Registered as `cls2yml`
"""
import sys
from typing import Dict

import argparse
import logging
from pathlib import Path

from allennlp.common import Registrable
from allennlp.commands.subcommand import Subcommand
from overrides import overrides
import re
import yaml

from allennlp_hydra.config.fill_defaults import (
    fill_config_with_default_values,
    get_positional_arguments,
)

logger = logging.getLogger(__name__)


@Subcommand.register("class2yaml")
class ClassToYaml(Subcommand):
    @overrides
    def add_subparser(
            self, parser: argparse._SubParsersAction
    ) -> argparse.ArgumentParser:
        description = """Convert a registered class to its YAML representation"""
        subparser = parser.add_parser(
            self.name, description=description, help=description
        )

        subparser.add_argument(
            "cls_name",
            type=str,
            help="Name the class was registered with. This"
                 " is will be used as the name of the "
                 "created `.yaml` file.",
        )

        subparser.add_argument(
            "base_cls_name", type=str, help="Name of the Base Class."
        )

        subparser.add_argument(
            "-s",
            "--serialization-dir",
            required=True,
            type=str,
            help="Directory to save the file to. The name of the file will "
                 "be `{cls_name}.json`. All non alphanumeric and `_` characters "
                 "will be replaced with `_`.",
        )

        subparser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help="Force overwriting of existing `.yaml` file.",
        )

        subparser.set_defaults(func=class_to_yaml_from_args)

        return subparser


def class_to_yaml_from_args(args: argparse.Namespace) -> Dict:
    return class_to_yaml(
        cls_name=args.cls_name,
        base_cls_name=args.base_cls_name,
        serialization_dir=args.serialization_dir,
        force=args.force,
    )


def class_to_yaml(
        cls_name: str, base_cls_name: str, serialization_dir: str, force: bool = False
) -> Dict:
    print(f"Looking for the '{base_cls_name}' registered as {cls_name}")
    base_class = None

    # Go through the registered classes and find the `base_cls_name`
    for base_class_object in Registrable.__subclasses__():
        if base_class_object.__name__ == base_cls_name:
            base_class = base_class_object
            break

    if base_class is None:
        raise ValueError(
            f"The class '{base_cls_name}' could not be found as a"
            f" subclass of Registrable."
        )

    try:
        cls_to_get = base_class.by_name(cls_name)
    except KeyError:
        raise KeyError(f"'{cls_name}' is not registered as a {base_cls_name}")

    config_for_class = fill_config_with_default_values(base_class, {"type": cls_name})

    # Add the positional args to the config.
    for postional_arg in get_positional_arguments(cls_to_get):
        # '???' is the OmegaConf variable for required. Will throw an error if
        # not filled.
        config_for_class[postional_arg] = "???"

    serialization_dir = Path(serialization_dir)
    if not serialization_dir.exists():
        raise ValueError(f"{serialization_dir} does not exist")

    file_name = re.sub(r"[^\w\._]+", "_", cls_name)
    out_path = serialization_dir.joinpath(f"{file_name}.yaml")
    if out_path.exists() and not force:
        raise ValueError(f"{out_path} already exists. use --force to override.")
    with out_path.open("w", encoding="utf-8") as out_file:
        yaml.dump(config_for_class, out_file, yaml.Dumper)

    print(f"Config for {base_cls_name}.{cls_name}:")
    print("")
    yaml.dump(config_for_class, sys.stdout, yaml.Dumper)

    return config_for_class



"""
The `compose` command creates an AllenNLP config by composing a set of `yaml` files with Hydra's
[`Compose API`](https://hydra.cc/docs/advanced/compose_api). Overriding
using Hydra's [`Override Grammar`](https://hydra.cc/docs/advanced/compose_api)
is also supported.

# Parameters

config_path: `Union[str, PathLike]`
    Path to the root config directory.

config_name: `str`
    The name of the root config file. Do NOT include the `.yaml`.

job_name: `str`
    The job name. This is passed to Hydra and is not used here.

-s/--serialization-dir: `Union[str, PathLike]`
    The directory where the new AllenNLP config will be saved to. The name used
     for saving will be the `config_name` and it will be a `.json` file.

-o/--overrides: `List[str]`, optional (default=`[]`)
    Keyword arguments passed will be used as a list of overrides using Hydra's
    override grammar for the config.

    Example usage:

    ```zsh
    allennlp compose conf config example -s example --overrides A=B C="D"
    ```
    Will be interpreted as overrides `['A=B', 'C="D"']`



# Example

Say you have the following directory structure:

```
project
+-- conf
|   +-- dataset_readers
|   |   +-- A.yaml
|   |   +-- B.yaml
|   +-- models
|   |   +-- C.yaml
|   |   +-- D.yaml
|   +-- config.yaml
+-- experiments
```

`conf/dataset_readers/A.yaml`:

```yml
type: A
start_token: <s>
end_token: </s>
```

`conf/dataset_readers/B.yaml`:

```yml
type: B
start_token: [CLS]
end_token: [SEP]
```


`conf/models/C.yaml`:

```
type: C
layers: 5
```

`conf/models/D.yaml`:

```YAML
type: D
input_dim: 10
```


`config.yaml`
```yml
defaults:
    - dataset_reader: A
    - model: C

debug: false
```

Then running the command
```zsh
allennlp compose conf config example -s experiments
```
Produces the file `project/experiments/config.json`
```json
{
    "dataset_reader":{
        "type": "A",
        "start_token": "<s>",
        "end_token": "</s>"
    },
    "model": {
        "type": "C",
        "layers": 5
    },
    "debug": false
}
```

If you want to override the config and use the `B` dataset reader with the `D`
model, you would modify the previous command:
```zsh
allennlp compose conf config example -s experiments -o model=D dataset_reader=B
```
Produces the file `project/experiments/config.json`
```json
{
    "dataset_reader":{
        "type": "B",
        "start_token": "[CLS]",
        "end_token": "[SEP]"
    },
    "model": {
        "type": "D",
        "input_dim": 10
    },
    "debug": false
}
```

And if you wanted to change `input_dim` of model `D` to 25:
```zsh
allennlp compose conf config example -s experiments -o model=D dataset_reader=B model.input_dim=25
```

Produces the file `project/experiments/config.json`
```json
{
    "dataset_reader":{
        "type": "B",
        "start_token": "[CLS]",
        "end_token": "[SEP]"
    },
    "model": {
        "type": "D",
        "input_dim": 25
    },
    "debug": false
}
```
"""

from typing import Dict, Union, List, Optional

import argparse
import json
from os import PathLike
import logging
from pathlib import Path

from allennlp.data import DataLoader, DatasetReader
from allennlp.training import Trainer
from allennlp.models import Model
from allennlp.commands.subcommand import Subcommand
import hydra
from omegaconf import OmegaConf
from overrides import overrides

from allennlp_hydra.config.fill_defaults import fill_config_with_default_values

logger = logging.getLogger(__name__)


@Subcommand.register("compose")
class ComposeConfig(Subcommand):
    @overrides
    def add_subparser(
        self, parser: argparse._SubParsersAction
    ) -> argparse.ArgumentParser:
        description = """Compose a config with Hydra"""
        subparser = parser.add_parser(
            self.name, description=description, help=description
        )

        subparser.add_argument(
            "config_path", type=str, help="Path to the config directory."
        )

        subparser.add_argument(
            "config_name", type=str, help="Name of the config file to use."
        )
        subparser.add_argument("job_name", type=str, help="Name of the job.")

        subparser.add_argument(
            "-s",
            "--serialization-dir",
            required=True,
            type=str,
            help="Directory to save the config to. The name of the config will "
            "be `{config_name}.json`",
        )
        subparser.add_argument(
            "-o",
            "--overrides",
            nargs="*",
            help="Any key=value arguments to override config values "
            "(use dots for.nested=overrides)",
        )

        subparser.add_argument(
            "--fill-defaults",
            action="store_true",
            default=False,
            help="Add default arguments from each loaded class to the config.",
        )

        subparser.set_defaults(func=compose_config_from_args)

        return subparser


def compose_config_from_args(args: argparse.Namespace) -> Dict:
    """
    Wrapper for compose so that it can be called with `argparse` arguments from
    the CLI.
    # Parameters
    args: `argparse.Namespace`
        The parsed args from `argparse`.

    # Returns
        The composed config.

    """
    return compose_config(
        config_path=args.config_path,
        config_name=args.config_name,
        job_name=args.job_name,
        serialization_dir=args.serialization_dir,
        config_overrides=args.overrides,
    )


def compose_config(
    config_path: Union[str, PathLike],
    config_name: str,
    job_name: str,
    serialization_dir: Optional[Union[str, PathLike]] = None,
    config_overrides: List[str] = None,
    fill_defaults: bool = False,
) -> Dict:
    """
    Create an AllenNLP config by composing a set of `yaml` files with Hydra's
    [`Compose API`](https://hydra.cc/docs/advanced/compose_api). Overriding
    using Hydra's [`Override Grammar`](https://hydra.cc/docs/advanced/compose_api)
    is also supported.

    # Parameters

    config_path: `Union[str, PathLike]`
        Path to the root config directory.

    config_name: `str`
        The name of the root config file.

    job_name: `str`
        The job name. This is passed to Hydra and is not used here.

    serialization_dir: `Optional[Union[str, PathLike]]`, optional (default=`None`)
        If this is passed, it is the directory where the new AllenNLP config
        will be saved to. The name used for saving will be the `config_name` and
         it will be a `.json` file.

    config_overrides: `List[str]`, optional (default=`[]`)
        List of overrides using Hydra's override grammar for the config.

    fill_defaults: `bool`, optional (default=`False`)
        Add arguments and their default values to the config if they are not
        specified.

    # Returns

    `Dict`
        The dictionary config generated by Hydra.

    """
    if config_overrides is None:
        config_overrides = []

    # Make the config path relative to the location of THIS file. I.E. make it
    # relative to the `allennlp_hydra/commands` subdirectory.
    config_path = Path(config_path).absolute().resolve()
    if not config_path.exists():
        raise ValueError(f"Config path '{config_path}' does not exist")
    if not config_path.is_dir():
        raise ValueError(f"Config path '{config_path}' is not a directory")

    # Compose the config with hydra.
    with hydra.initialize_config_dir(config_dir=str(config_path), job_name=job_name):
        cfg = hydra.compose(config_name=config_name, overrides=config_overrides)

    # cfg is a `DictConfig` object, so we need to convert it to a normal dict
    # using OmegaConf in order to save it.
    cfg = OmegaConf.to_container(cfg, resolve=True)

    # If filling the defaults, fill them here.
    if fill_defaults:
        cfg["data_loader"] = fill_config_with_default_values(
            DataLoader, cfg["data_loader"]
        )
        cfg["dataset_reader"] = fill_config_with_default_values(
            DatasetReader, cfg["dataset_reader"]
        )
        cfg["model"] = fill_config_with_default_values(Model, cfg["model"])
        cfg["trainer"] = fill_config_with_default_values(Trainer, cfg["trainer"])

    # We only save if a serialization dir was passed.
    if serialization_dir is not None:
        cfg_save_path = Path(serialization_dir).joinpath(f"{config_name}.json")
        with cfg_save_path.open("w", encoding="utf-8") as cfg_file:
            # Add the extra options for readability.
            json.dump(cfg, cfg_file, indent=True, sort_keys=True)

    return cfg



"""
Taken from
https://github.com/allenai/allennlp-models/blob/main/allennlp_models/version.py
"""
_MAJOR = "0"
_MINOR = "0"
_PATCH = "7"

VERSION_SHORT = f"{_MAJOR}.{_MINOR}"
VERSION = f"{VERSION_SHORT}.{_PATCH}"
__version__ = VERSION



import setuptools
from allennlp_hydra.version import VERSION

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="allennlp-hydra",
    version=VERSION,
    author="Gabriel Orlanski",
    author_email="gabeorlanski@gmail.com",
    description="Hydra plugin for allennlp",
    license="Apache 2.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "test_*"],
    ),
    project_urls={
        "Documentation": "https://github.com/gabeorlanski/allennlp-hydra",
        "Source Code": "https://github.com/gabeorlanski/allennlp-hydra",
    },
    install_requires=[
        "allennlp>=2.6.0",
        "hydra-core>=1.1.1",
        "overrides>=3.1.0",
        "omegaconf>=2.1",
    ],
    requires_python=">=3.7",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



"""
Module for adding arguments and their default values to a config.
"""
from typing import Dict, Union, Any, get_args, Iterable

from copy import deepcopy
import inspect
import logging

from allennlp.common import Registrable, Params, Lazy, FromParams

logger = logging.getLogger(__name__)


def fill_config_with_default_values(
        base_class: Union[FromParams, Registrable], config: Dict
) -> Dict:
    """
    Fill a `config` with the arguments and their default values from a
    `base_class`. When it encounters nested objects (i.e. non-builtin classes,
    for now limited to AllenNLP `FromParams` and `Registrable`) it will recurse
    and create the dictionary with defaults.

    This does NOT change the mutable config nor does it override keys if they
    exist in the config.

    **NOTE:** When AllenNLP expands upon their
    [`to_params`](https://docs.allennlp.org/main/api/common/from_params/#to_params)
    api, this will rely on that, but for the time being it relies on the
    [`inspect`](https://docs.python.org/3/library/inspect.html) module.

    # Parameters
    base_class: `Union[FromParams, Registrable]`
        The base class that you want to use to fill the `config` with the
        arguments and their default values. These arguments are for the
        `base_class.__init__` function.
    config: `Dict`
        The configuration dict to fill with the defaults. It does **NOT** make
        any changes to the mutable data but instead uses `deepcopy` to copy it
        to a new object.
        Existing keys will not be overwritten.
    # Returns
    cfg_with_defaults: `Dict` The filled config with defaults.
    """
    if isinstance(config, Params):
        raise TypeError("Params object cannot be passed to fill defaults")

    # Copy to avoid mutable changes
    config_dict = deepcopy(config)

    # This is a band-aid hack for when there are complex type annotations that
    # are too difficult to handle (Nested dictionaries), we instead just return
    # the config as is.
    if base_class is None:
        return config_dict

    # Check if we need to use the default or not.
    if issubclass(base_class, Registrable):
        class_to_use = base_class.by_name(
            config_dict.get("type", base_class.default_implementation)
        )
    else:
        class_to_use = base_class

    # Get the init parameters from the underlying class of `initialized_class`
    init_params = inspect.signature(class_to_use)

    output_config = deepcopy(config_dict)
    for parameter in init_params.parameters.values():
        # Positional parameters do not have default values, so we skip them.
        if parameter.default == inspect.Parameter.empty:
            continue

        # The parameter is a keyword and has a default value.
        if parameter.name not in config_dict:
            output_config[parameter.name] = get_default_value_for_parameter(parameter)
            continue

        if parameter.annotation is None:
            raise ValueError(
                f"Argument {parameter.name} in class"
                f" {base_class.__name__} does not have a type"
                f" annotation"
            )

        # If it is not a dict, then it is not a class constructor argument and
        # thus no need to recurse.
        if not isinstance(config_dict[parameter.name], dict):
            continue

        output_config[parameter.name] = dict(
            fill_config_with_default_values(
                get_annotation_class(parameter), config_dict[parameter.name]
            )
        )

    return output_config


def get_default_value_for_parameter(parameter: inspect.Parameter) -> Any:
    """
    Get the default value for a parameter.

    If there is a type annotation, check if it is a subclass of Registrable. If
    that is the case, check if the default value is an instance of that class.

    If it is, then recurse. Otherwise continue. Handles the case where the
    default value is an initialized registrable. We also find what name the
    class was registered as to add that as the `type` value.

    # Parameters
        parameter: `inspect.Parameter`
            The parameter to get the default value of.

    # Returns
    `Any` The default value of the parameter.
    """
    if parameter.annotation is not None:
        # Annotation found, check if it is a Registrable.
        annotation_type = get_annotation_class(parameter)
        if annotation_type is None:
            return parameter.default

        try:
            is_subclass_of_registrable = issubclass(annotation_type, Registrable)
        except TypeError:
            # If there is some typing error, return the default value and call
            # it a day.
            return parameter.default

        cfg_dict = None
        if is_subclass_of_registrable and isinstance(
                parameter.default, annotation_type
        ):
            # Reverse lookup what the class was registered as. We do need to
            # access the protected `_registry` attribute of registrable, but
            # there is currently no other way to do this.
            parameter_default_class = parameter.default.__class__
            registered_name = None
            if annotation_type not in Registrable._registry:
                raise KeyError(
                    f"'{parameter.annotation.__name__}' has no" f" registered classes"
                )

            for name, registered_class in Registrable._registry[
                parameter.annotation
            ].items():
                if registered_class[0] == parameter_default_class:
                    registered_name = name
                    break

            if registered_name is None:
                raise KeyError(
                    f"'{parameter_default_class.__name__}' was never" f" registered."
                )

            cfg_dict = {"type": registered_name}
        elif is_subclass_of_registrable and isinstance(parameter.default, Lazy):
            logger.warning(
                f"{parameter.name} has a Lazy object for its "
                f"default. That is not currently supported and will"
                f" be handled as getting the arguments for the"
                f" default annotation."
            )

            cfg_dict = {"type": annotation_type.default_implementation}
        elif not is_subclass_of_registrable and issubclass(annotation_type, FromParams):
            cfg_dict = {}

        if cfg_dict is not None:
            return fill_config_with_default_values(annotation_type, cfg_dict)

    # No annotation found or it is not a registrable, so return the default
    # value.
    return parameter.default


def get_annotation_class(parameter: inspect.Parameter) -> Union[type, None]:
    """
    Get the class of the annotation.
    # Parameters
        parameter: `inspect.Parameter`
            The parameter to get the class of.
    # Returns
    `Union[type, None]` The type or None.
    """
    # If the type is a `_GenericAlias`, we care what the arguments are inside.
    # Thus, we get the annotation args. If it is empty, it is not a
    # `_GenericAlias` and we can return the value.
    annotation_args = get_args(parameter.annotation)
    if not annotation_args:
        return parameter.annotation

    # If there are more than one argument, we return None because it is easier
    # to handle.
    if len(annotation_args) > 1:
        return None

    # Otherwise return the found type.
    return annotation_args[0]


def get_positional_arguments(cls_type: FromParams) -> Iterable[str]:
    # Get the init parameters from the underlying class of `initialized_class`
    init_params = inspect.signature(cls_type)

    for parameter in init_params.parameters.values():
        if (
                parameter.default == inspect.Parameter.empty
                and parameter.kind != inspect.Parameter.VAR_KEYWORD
        ):
            yield parameter.name
